#include <oml2/omlc.h>
#include <sys/time.h>
#include <stdio.h>
#include <pthread.h> 
#include <errno.h>
#include <sys/socket.h> /* for socket(), bind(), and connect() */
#include <arpa/inet.h>  /* for sockaddr_in and inet_ntoa() */
#include <stdlib.h>     /* for atoi() and exit() */
#include <string.h>     /* for memset() */
#include <unistd.h>     /* for close() */

#define MAXPENDING 5    /* Maximum outstanding connection requests */
#define SIZE 30
#define ECHOMAX 255     /* Longest string to ping */

void DieWithError(char *errorMessage);  /* Error handling function */


unsigned short MR_Port; /* Port on which pigeon's real Messages are Recived*/
unsigned short echoServPort; /* Port on which pings are received */
unsigned short MT_Port; /* Port from which real Messages are Transfered to pigeon*/

//following  3 semaphores are to handle enqueue and dequeue in Queue_Fs 

pthread_mutex_t region_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t space_available = PTHREAD_COND_INITIALIZER;
pthread_cond_t data_available = PTHREAD_COND_INITIALIZER;

int Queue_Fs[SIZE]; // F stands for Foreign and s for sink 
int size = 0;  /* number of full elements */
int front,rear=0;  /* queue */

//following  3 semaphores are to handle enqueue and dequeue in Queue_Fg 
pthread_mutex_t region_mutex1 = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t space_available1 = PTHREAD_COND_INITIALIZER;
pthread_cond_t data_available1 = PTHREAD_COND_INITIALIZER;

int Queue_Fg[SIZE]; // F stands for Foreign and g for generation 
int size1 = 0;  /* number of full elements */
int front1,rear1=0;  /* queue */

int generationDelay=1; // default message generation delay
time_t t;// for seed of the random number generator

// following 1 semaphore to ensure mutual exclusion among omlc_inject() calls

pthread_mutex_t my_oml_mutex = PTHREAD_MUTEX_INITIALIZER; 

struct message{
              int id;
              char data[10000000];
       }; // we use this structure to realize a big message





// Below is OML measurement point
OmlMPDef mp_def [] =
{
  { "msgID", OML_LONG_VALUE },
  { "tag", OML_STRING_VALUE },
  { NULL, (OmlValueT)0 }
};
// some oml definition which is later used
OmlMP* mp;



 
main(int argc, const char* argv[])  
{
  pthread_t MCM_thread; // Thread running MCM (Message Consumption Module) 
  pthread_t MRM_thread; // Thread running MRM (Message Receiving Module)
  pthread_t CRM_thread;// Thread running CRM (Connectivity Response Module)
  pthread_t MGM_thread; // Thread running MGM (Message Generation Module) 
  pthread_t MTM_thread; // Thread running MTM (Message Transfer Module)

  void *MCM();
  void *MRM();
  void *CRM();
  void *MGM();
  void *MTM();
 

// OML init now

  int result = omlc_init ("ForeignOmlClient", &argc, argv, NULL);
  if (result == -1) {
    fprintf (stderr, "Could not initialize OML\n");
    exit (1);
  }
// OML init ends



 
   if (argc != 4 && argc != 5)     /* Test for correct number of arguments */
     {
         fprintf(stderr, "Usage: %s <MessageTransferPort> <MessageReceiptPort> <Ping Port> [<generationDelay-in-seconds(Integer)>] \n", argv[0]);
         exit(1);
     }

   MT_Port = atoi(argv[1]);  /* First arg:  local message transfer port */
   MR_Port = atoi(argv[2]);  /* Second arg:  local message receipt port */

   echoServPort = atoi(argv[3]); /* Third arg: local ping port */
   if(argc==5)
     generationDelay=atoi(argv[4]); // Fourth arg: genration delay

   srand((unsigned)2); // replace 2 by (time(&t)); CHECK


  // ROY adds OML MP

  mp = omlc_add_mp ("infoForeign", mp_def);

  if (mp == NULL) {
    fprintf (stderr, "Error: could not register Measurement Point \"info\"");
    exit (1);
  }

 // OML omlc add up ends

 // OML omlc_start
  result = omlc_start();

  if (result == -1) {
    fprintf (stderr, "Error starting up OML measurement streams\n");
    exit (1);
  }




    pthread_create(&MTM_thread,NULL,MTM,NULL);
    pthread_create(&MGM_thread,NULL,MGM,NULL);
    pthread_create(&CRM_thread, NULL, CRM, NULL);
    pthread_create(&MCM_thread,NULL,MCM,NULL);
    pthread_create(&MRM_thread,NULL,MRM,NULL);
    pthread_join(MTM_thread,NULL);

    omlc_close();  // ROY CHECK if it is the right place for omlc_close()

}

void add_queue(int i){
  Queue_Fs[rear] = i;
  rear = (rear+1) % SIZE;
  size++;
}

int get_queue(){
  int v;
  v = Queue_Fs[front];
  front= (front+1) % SIZE;
  size--;
  return v ;
}


void add_queue1(int i){
  Queue_Fg[rear1] = i;
  rear1 = (rear1+1) % SIZE;
  size1++;
}

int get_queue1(){
  int v;
  v = Queue_Fg[front1];
  front1= (front1+1) % SIZE;
  size1--;
  return v ;
}

 
void *MCM()
{
  int v = 0;
  while (1) {
    pthread_mutex_lock(&region_mutex);
    if (size == 0) {
       printf("MCM: sink queue is empty\n");
       pthread_cond_wait(&data_available,&region_mutex);
     }

    v = get_queue();
    pthread_cond_signal(&space_available);
    pthread_mutex_unlock(&region_mutex);

  
    printf("consuming at Foreign host  %d H \n",v);
  } // while(1)

  pthread_exit(NULL);
}



void *MRM()
{
    int v;

    int servSock;                    /* Socket descriptor for server */
    int clntSock;                    /* Socket descriptor for client */
    struct sockaddr_in ServAddr; /* Local address */
    struct sockaddr_in ClntAddr; /* Client address */
    unsigned int clntLen;            /* Length of client address data structure */

    int number, ByteRcvd; // ROY add
    int j; // ROY add    
    struct message buffer; 

    /* Create socket for incoming connections */
    if ((servSock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
        DieWithError("socket() failed");
      
    /* Construct local address structure */
    memset(&ServAddr, 0, sizeof(ServAddr));   /* Zero out structure */
    ServAddr.sin_family = AF_INET;                /* Internet address family */
    ServAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* Any incoming interface */
    ServAddr.sin_port = htons(MR_Port);      /* Local port */

    /* Bind to the local address */
    if (bind(servSock, (struct sockaddr *) &ServAddr, sizeof(ServAddr)) < 0)
        DieWithError("bind() failed");

    /* Mark the socket so it will listen for incoming connections */
    if (listen(servSock, MAXPENDING) < 0)
        DieWithError("listen() failed");

    for (;;) /* Run forever */
    {
        /* Set the size of the in-out parameter */
        clntLen = sizeof(ClntAddr);

        /* Wait for a client to connect */
        if ((clntSock = accept(servSock, (struct sockaddr *) &ClntAddr, 
                               &clntLen)) < 0)
            DieWithError("accept() failed");

        /* clntSock is connected to a client! */

        printf("Handling pigeon  %s\n", inet_ntoa(ClntAddr.sin_addr));

        
        /* Receive message from client */
        if ((ByteRcvd = recv(clntSock, &number, sizeof(int), 0)) < 0)
           DieWithError("recv() failed");
        printf("%d=number and %d=numberOfBytes Received \n", number, ByteRcvd);

        for(j=0;j<number;j++)
           {

             pthread_mutex_lock(&region_mutex);
             if (size == SIZE) {

             printf("****the sink queue is full**** \n");
             pthread_cond_wait(&space_available,&region_mutex); 
             }


           if (recv(clntSock, &buffer, sizeof(struct message), MSG_WAITALL) != sizeof(struct message))
               DieWithError("recv2() failed");
 
           printf("size of received message is %d \n", sizeof(struct message));
           v = buffer.id; 
           add_queue(v);
           pthread_cond_signal(&data_available);
           pthread_mutex_unlock(&region_mutex);
           printf("receved from pigeon %d  \n",v);



        // oml injecting some data


       long msg_id = (long)v;
       char* desc ="ReceivedAtForeign";

       OmlValueU values[2];

       omlc_set_long (values[0], msg_id);
       omlc_set_string (values[1], desc);

      pthread_mutex_lock(&my_oml_mutex);
     // lock oml mutex because omlc_inject is not thread-safe

      omlc_inject (mp, values);

     pthread_mutex_unlock(&my_oml_mutex); // unlock oml mutex


      // oml injection ends




           } // for()

     close(clntSock);    /* Close client socket */        

     } // for(;;)
pthread_exit(NULL);
}


void *CRM()
{
    int sock;                        /* Socket */
    struct sockaddr_in echoServAddr; /* Local address */
    struct sockaddr_in echoClntAddr; /* Client address */
    unsigned int cliAddrLen;         /* Length of incoming message */
    char echoBuffer[ECHOMAX];        /* Buffer for echo string */
    int recvMsgSize;                 /* Size of received message */

    /* Create socket for sending/receiving datagrams */
    if ((sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
        DieWithError("socket() failed");

    /* Construct local address structure */
    memset(&echoServAddr, 0, sizeof(echoServAddr));   /* Zero out structure */
    echoServAddr.sin_family = AF_INET;            /* Internet address family */
    echoServAddr.sin_addr.s_addr = htonl(INADDR_ANY);/*Any incoming interface */
    echoServAddr.sin_port = htons(echoServPort);      /* Local port */

    /* Bind to the local address */
    if (bind(sock, (struct sockaddr *) &echoServAddr, sizeof(echoServAddr)) < 0)
        DieWithError("bind() failed");
 
    for (;;) /* Run forever */
      {
         /* Set the size of the in-out parameter */
         cliAddrLen = sizeof(echoClntAddr);

         /* Block until receive message from a client */
         if ((recvMsgSize = recvfrom(sock, echoBuffer, ECHOMAX, 0,
            (struct sockaddr *) &echoClntAddr, &cliAddrLen)) < 0)
            DieWithError("recvfrom() failed");

         printf("Handling ping from client %s\n", inet_ntoa(echoClntAddr.sin_addr));
         printf("ping received from pigeon is %s \n", echoBuffer);
         /* Send received datagram back to the client */
         if (sendto(sock, echoBuffer, recvMsgSize, 0, (struct sockaddr *) &echoClntAddr, sizeof(echoClntAddr)) != recvMsgSize)
            DieWithError("sendto() sent a different number of bytes than expected");
     }


    pthread_exit(NULL);
}

void *MGM()
 {
  int i = 0;
  int j=0;
  while (1) {

    j=rand()%(2*generationDelay*1000); // average generation delay = generationDelay (SECONDS); note that j is in millisecodns 
    usleep(j*1000); // sleeps for j milliseconds. note that usleep for microsec

  //******** ROY ***should we move the following oml code after  add_queue(i) ? 

  // oml injecting some data


  long msg_id = (long)i;
  char* desc ="generatedAtForeign";

  OmlValueU values[2];

  omlc_set_long (values[0], msg_id);
  omlc_set_string (values[1], desc);


      pthread_mutex_lock(&my_oml_mutex);
     // lock oml mutex because omlc_inject is not thread-safe

      omlc_inject (mp, values);

     pthread_mutex_unlock(&my_oml_mutex); // unlock oml mutex


    // oml injection ends


    pthread_mutex_lock(&region_mutex1);
    if (size1 == SIZE) {
       printf("MGM: generation queue is full \n");
       pthread_cond_wait(&space_available1,&region_mutex1);
   }
  add_queue1(i);
  printf("MGM add_queue %d F\n",i);
  pthread_cond_signal(&data_available1);
  pthread_mutex_unlock(&region_mutex1);
  i = i + 1;
 }
 pthread_exit(NULL);
}

void *MTM()
{
    int i,v;

    int servSock;                    /* Socket descriptor for server */
    int clntSock;                    /* Socket descriptor for client */
    struct sockaddr_in ServAddr; /* Local address */
    struct sockaddr_in ClntAddr; /* Client address */
    unsigned int clntLen;            /* Length of client address data structure */

    int number, numRcvd; // ROY add
    int j; // ROY add    



    struct message buffer;
    memset(&buffer, 0, sizeof(struct message));
    buffer.id = -1;



    /* Create socket for incoming connections */
    if ((servSock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
        DieWithError("socket() failed");

    /* Construct local address structure */
    memset(&ServAddr, 0, sizeof(ServAddr));   /* Zero out structure */
    ServAddr.sin_family = AF_INET;                /* Internet address family */

    ServAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* Any incoming interface */
    ServAddr.sin_port = htons(MT_Port);      /* Local port */

    /* Bind to the local address */
    if (bind(servSock, (struct sockaddr *) &ServAddr, sizeof(ServAddr)) < 0)
        DieWithError("bind() failed");

    /* Mark the socket so it will listen for incoming connections */
    if (listen(servSock, MAXPENDING) < 0)
        DieWithError("listen() failed");

    for (;;) /* Run forever */
    {
        /* Set the size of the in-out parameter */
        clntLen = sizeof(ClntAddr);

        /* Wait for a client to connect */
        if ((clntSock = accept(servSock, (struct sockaddr *) &ClntAddr,
                               &clntLen)) < 0)
            DieWithError("accept() failed");
       /* clntSock is connected to a client! */

        printf("Handling client %s\n", inet_ntoa(ClntAddr.sin_addr));

        number = size1; // ROY needs to check for atomicity    

      // ROY needs to change the name of numRecvd to ByteSent below   

       if ((numRcvd = send(clntSock, &number, sizeof(int), 0)) < 0)
           DieWithError("send() failed");
      // ROY needs to CHECK the following readjustment of number

       if ((numRcvd = recv(clntSock, &number, sizeof(int), 0)) < 0)
           DieWithError("recv() failed");

       printf("%d=number and %d=numberOfBytes \n", number, numRcvd);

       for(j=0;j<number;j++)
        {

          pthread_mutex_lock(&region_mutex1);
          if (size1 == 0) {
            printf("***foreign host message generation queue is empty*** \n");
            pthread_cond_wait(&data_available1,&region_mutex1);
          }
          v = get_queue1();
          pthread_cond_signal(&space_available1);
          pthread_mutex_unlock(&region_mutex1);
          printf("sending to pigeon %d  \n",v);

          buffer.id =v;

          if (send(clntSock, &buffer, sizeof(struct message), 0) != sizeof(struct message))
             DieWithError("send() failed");

         printf("size of delivered message is %d \n", sizeof(struct message));

     // oml injecting some data


    long msg_id = (long)v;
    char* desc ="TransferredByForeign";

    OmlValueU values[2];

    omlc_set_long (values[0], msg_id);
    omlc_set_string (values[1], desc);


      pthread_mutex_lock(&my_oml_mutex);
     // lock oml mutex because omlc_inject is not thread-safe

      omlc_inject (mp, values);

     pthread_mutex_unlock(&my_oml_mutex); // unlock oml mutex


    // oml injection ends


        } // for()

     close(clntSock);    /* Close client socket */

     } // for(;;)
 pthread_exit(NULL);
}

