#include <sys/time.h>
#include <stdio.h>
#include <pthread.h> 
#include <errno.h>
#include <sys/socket.h> /* for socket(), bind(), and connect() */
#include <arpa/inet.h>  /* for sockaddr_in and inet_ntoa() */
#include <stdlib.h>     /* for atoi() and exit() */
#include <string.h>     /* for memset() */
#include <unistd.h>     /* for close() */
#include <time.h>

#define MAXPENDING 5    /* Maximum outstanding connection requests */
#define SIZE 30
#define ECHOMAX 255     /* Longest string to ping */

unsigned short MT_Port; /* Port on which real Messages are Transfered to pigeon*/
unsigned short echoServPort; /* Port on which pings are received */
unsigned short MR_Port; /* Port on which real Messages are Received from pigeon*/

void DieWithError(char *errorMessage);  /* Error handling function */

// following 3 semaphores handles enqueue and dequeue of Queue_Hg

pthread_mutex_t region_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t space_available = PTHREAD_COND_INITIALIZER;
pthread_cond_t data_available = PTHREAD_COND_INITIALIZER;

int Queue_Hg[SIZE]; // H stands for Home and g for generation 
int size = 0;  /* number of full elements */
int front,rear=0;  /* queue */


// following 3 semaphores handles enqueue and dequeue of Queue_Hs

pthread_mutex_t region_mutex1 = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t space_available1 = PTHREAD_COND_INITIALIZER;
pthread_cond_t data_available1 = PTHREAD_COND_INITIALIZER;

int Queue_Hs[SIZE]; // H stands for Home and s for sink
int size1 = 0;  /* number of full elements */
int front1,rear1=0;  /* queue */


int generationDelay=1; // default message generation delay
time_t t;// for seed of the random number generator
 
main(int argc, char* argv[])  
{
  pthread_t MGM_thread; // Thread running MGM (Message Generation Module) 
  pthread_t MTM_thread; // Thread running MTM (Message Transfer Module)
  pthread_t CRM_thread;// Thread running CRM (Connectivity Response Module)
  pthread_t MCM_thread; // Thread running MCM (Message Consumption Module) 
  pthread_t MRM_thread; // Thread running MRM (Message Receiving Module)


  void *MGM();
  void *MTM();
  void *CRM();
  void *MCM();
  void *MRM();
 
  if (argc != 4 && argc != 5)     /* Test for correct number of arguments */
    {
        fprintf(stderr, "Usage: %s <MessageTransferPort> <MessageReceiptPort> <Ping Port> [<generationDelay-in-seconds(Integer)>] \n", argv[0]);
        exit(1);
    }
  MT_Port = atoi(argv[1]);  /* First arg:  local message transfer port */
  MR_Port = atoi(argv[2]);  /* Second arg:  local message receipt port */

  echoServPort = atoi(argv[3]); /* Third arg: local ping port */
  if(argc==5) 
    generationDelay=atoi(argv[4]); // Fourth arg: genration delay

  srand((unsigned)2); // replace 2 by (time(&t)); CHECK


    pthread_create(&MTM_thread,NULL,MTM,NULL);
    pthread_create(&MGM_thread,NULL,MGM,NULL);
    pthread_create(&CRM_thread, NULL, CRM, NULL);
    pthread_create(&MCM_thread,NULL,MCM,NULL);
    pthread_create(&MRM_thread,NULL,MRM,NULL);
    pthread_join(MTM_thread,NULL);

}

void add_queue(int i){
  Queue_Hg[rear] = i;
  rear = (rear+1) % SIZE;
  size++;
}

int get_queue(){
  int v;
  v = Queue_Hg[front];
  front= (front+1) % SIZE;
  size--;
  return v ;
}


void add_queue1(int i){
  Queue_Hs[rear1] = i;
  rear1 = (rear1+1) % SIZE;
  size1++;
}

int get_queue1(){
  int v;
  v = Queue_Hs[front1];
  front1= (front1+1) % SIZE;
  size1--;
  return v ;
}
 

 
void *MGM()
 {
  int i = 0;
  int j=0;
  while (1) {

    j=rand()%(2*generationDelay); // average generation delay = generationDelay 
    sleep(j);
    pthread_mutex_lock(&region_mutex);
    if (size == SIZE) {
       printf("MGM: generation queue is full \n");
       pthread_cond_wait(&space_available,&region_mutex);
   }
  add_queue(i);
  printf("MGM add_queue %d H \n",i);
  pthread_cond_signal(&data_available);
  pthread_mutex_unlock(&region_mutex);
  i = i + 1;
 }
 pthread_exit(NULL);
}


void *MTM()
{
    int i,v;

    int servSock;                    /* Socket descriptor for server */
    int clntSock;                    /* Socket descriptor for client */
    struct sockaddr_in ServAddr; /* Local address */
    struct sockaddr_in ClntAddr; /* Client address */
    unsigned int clntLen;            /* Length of client address data structure */

    int number, numRcvd; // ROY add
    int j; // ROY add    


    /* Create socket for incoming connections */
    if ((servSock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
        DieWithError("socket() failed");
      
    /* Construct local address structure */
    memset(&ServAddr, 0, sizeof(ServAddr));   /* Zero out structure */
    ServAddr.sin_family = AF_INET;                /* Internet address family */
    ServAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* Any incoming interface */
    ServAddr.sin_port = htons(MT_Port);      /* Local port */

    /* Bind to the local address */
    if (bind(servSock, (struct sockaddr *) &ServAddr, sizeof(ServAddr)) < 0)
        DieWithError("bind() failed");

    /* Mark the socket so it will listen for incoming connections */
    if (listen(servSock, MAXPENDING) < 0)
        DieWithError("listen() failed");

    for (;;) /* Run forever */
    {
        /* Set the size of the in-out parameter */
        clntLen = sizeof(ClntAddr);

        /* Wait for a client to connect */
        if ((clntSock = accept(servSock, (struct sockaddr *) &ClntAddr, 
                               &clntLen)) < 0)
            DieWithError("accept() failed");

        /* clntSock is connected to a client! */

        printf("Handling client %s\n", inet_ntoa(ClntAddr.sin_addr));

       number = size; // ROY needs to check for atomicity    

   // ROY needs to change the name of numRecvd to numSent below   
 
    if ((numRcvd = send(clntSock, &number, sizeof(int), 0)) < 0)
        DieWithError("send() failed");
// ROY needs to CHECK the following readjustment of number
   
    if ((numRcvd = recv(clntSock, &number, sizeof(int), 0)) < 0)
        DieWithError("recv() failed");

    printf("%d=number and %d=numberOfBytes \n", number, numRcvd);

    for(j=0;j<number;j++)
    {

     pthread_mutex_lock(&region_mutex);
     if (size == 0) {
         printf("***home host message generation queue is empty*** \n");
         pthread_cond_wait(&data_available,&region_mutex); 
     }
     v = get_queue();
     pthread_cond_signal(&space_available);
     pthread_mutex_unlock(&region_mutex);
     printf("sending to pigeon %d  \n",v);


       if (send(clntSock, &v, sizeof(int), 0) != sizeof(int))
            DieWithError("send() failed"); 
    } // for()

     close(clntSock);    /* Close client socket */        

     } // for(;;)
pthread_exit(NULL);
}

void *CRM()
{
    int sock;                        /* Socket */
    struct sockaddr_in echoServAddr; /* Local address */
    struct sockaddr_in echoClntAddr; /* Client address */
    unsigned int cliAddrLen;         /* Length of incoming message */
    char echoBuffer[ECHOMAX];        /* Buffer for echo string */
    int recvMsgSize;                 /* Size of received message */

    /* Create socket for sending/receiving datagrams */
    if ((sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
        DieWithError("socket() failed");

    /* Construct local address structure */
    memset(&echoServAddr, 0, sizeof(echoServAddr));   /* Zero out structure */
    echoServAddr.sin_family = AF_INET;            /* Internet address family */
    echoServAddr.sin_addr.s_addr = htonl(INADDR_ANY);/*Any incoming interface */
    echoServAddr.sin_port = htons(echoServPort);      /* Local port */


  /* Bind to the local address */
    if (bind(sock, (struct sockaddr *) &echoServAddr, sizeof(echoServAddr)) < 0)
        DieWithError("bind() failed");

    for (;;) /* Run forever */
      {
         /* Set the size of the in-out parameter */
         cliAddrLen = sizeof(echoClntAddr);

         /* Block until receive message from a client */
         if ((recvMsgSize = recvfrom(sock, echoBuffer, ECHOMAX, 0,
            (struct sockaddr *) &echoClntAddr, &cliAddrLen)) < 0)
            DieWithError("recvfrom() failed");

         printf("Handling ping from client %s\n", inet_ntoa(echoClntAddr.sin_addr));
         printf("ping received from pigeon is %s \n", echoBuffer);
         /* Send received datagram back to the client */
         if (sendto(sock, echoBuffer, recvMsgSize, 0, (struct sockaddr *) &echoClntAddr, sizeof(echoClntAddr)) != recvMsgSize)
            DieWithError("sendto() sent a different number of bytes than expected");
     }


  pthread_exit(NULL);
}

void *MCM()
{
  int v = 0;
  while (1) {
    pthread_mutex_lock(&region_mutex1);
    if (size1 == 0) {
       printf("MCM: sink queue is empty\n");
       pthread_cond_wait(&data_available1,&region_mutex1);
     }

    v = get_queue1();
    pthread_cond_signal(&space_available1);
    pthread_mutex_unlock(&region_mutex1);


    printf("consuming at Home host  %d F \n",v);
  } // while(1)

  pthread_exit(NULL);
}


void *MRM()
{
    int v;

    int servSock;                    /* Socket descriptor for server */
    int clntSock;                    /* Socket descriptor for client */
    struct sockaddr_in ServAddr; /* Local address */
    struct sockaddr_in ClntAddr; /* Client address */
    unsigned int clntLen;            /* Length of client address data structure */

    int number, ByteRcvd; // ROY add
    int j; // ROY add    


    /* Create socket for incoming connections */
    if ((servSock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
        DieWithError("socket() failed");

    /* Construct local address structure */
    memset(&ServAddr, 0, sizeof(ServAddr));   /* Zero out structure */
    ServAddr.sin_family = AF_INET;                /* Internet address family */
    ServAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* Any incoming interface */
    ServAddr.sin_port = htons(MR_Port);      /* Local port */


   /* Bind to the local address */
    if (bind(servSock, (struct sockaddr *) &ServAddr, sizeof(ServAddr)) < 0)
        DieWithError("bind() failed");

    /* Mark the socket so it will listen for incoming connections */
    if (listen(servSock, MAXPENDING) < 0)
        DieWithError("listen() failed");

    for (;;) /* Run forever */
    {
        /* Set the size of the in-out parameter */
        clntLen = sizeof(ClntAddr);

        /* Wait for a client to connect */
        if ((clntSock = accept(servSock, (struct sockaddr *) &ClntAddr,
                               &clntLen)) < 0)
            DieWithError("accept() failed");

        /* clntSock is connected to a client! */

        printf("from MRM(): Handling pigeon  %s\n", inet_ntoa(ClntAddr.sin_addr));


       /* Receive message from client */
        if ((ByteRcvd = recv(clntSock, &number, sizeof(int), 0)) < 0)
           DieWithError("recv() failed");
        printf("%d=number and %d=numberOfBytes Received \n", number, ByteRcvd);

        for(j=0;j<number;j++)
           {

             pthread_mutex_lock(&region_mutex1);
             if (size1 == SIZE) {

             printf("****the sink queue is full**** \n");
             pthread_cond_wait(&space_available1,&region_mutex1);
             }


           if (recv(clntSock, &v, sizeof(int), 0) != sizeof(int))
               DieWithError("recv2() failed");

           add_queue1(v);
           pthread_cond_signal(&data_available1);
           pthread_mutex_unlock(&region_mutex1);
           printf("receved from pigeon %d  \n",v);


           } // for()

     close(clntSock);    /* Close client socket */

     } // for(;;)
pthread_exit(NULL);

}

