#include <sys/time.h>
#include <stdio.h>
#include <pthread.h> 

#include <sys/socket.h> /* for socket(), connect(), send(), and recv() */
#include <arpa/inet.h>  /* for sockaddr_in and inet_addr() */
#include <stdlib.h>     /* for atoi() and exit() */
#include <string.h>     /* for memset() */
#include <unistd.h>     /* for close() */
#include <time.h>
#include <errno.h>      /* for errno */

#define  ECHOMAX         255     /* Longest string for ping echo */
#define TIMEOUT_SECS    3       /* Seconds between retransmits */
#define MAXTRIES        5       /* Tries before giving up */
#define PING_PORT     31000     /* ping echo port */


int C( char* echoServIP, unsigned short echoServPort); // this function checks the connectivity of the Pigeon with Home Host or the Foreign Host.

int SIZE_PF; // size of the queue which stores messages destined for Foreign Host

int SIZE_PH; // size of the queue which stores messages destined for Home Host

void DieWithError(char *errorMessage);  /* Error handling function */



char HomeIP[10]; // IP of home host
unsigned short  HomeMT_Port; // message transfer port of home host
unsigned short  HomeMR_Port; // message receiving port of home host

char foreignIP[10]; // IP of foreign host
unsigned short  foreignMR_Port; // message receiving port of foreign host
unsigned short  foreignMT_Port; // message transfer port of foreign host

//3 semaphores to ensure atomicity for Queue_PF

pthread_mutex_t region_mutex = PTHREAD_MUTEX_INITIALIZER; 
pthread_cond_t space_available = PTHREAD_COND_INITIALIZER;
pthread_cond_t data_available = PTHREAD_COND_INITIALIZER;

//3 semaphores to ensure atomicity for Queue_PH

pthread_mutex_t region_mutex1 = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t space_available1 = PTHREAD_COND_INITIALIZER;
pthread_cond_t data_available1 = PTHREAD_COND_INITIALIZER;

int* Queue_PF; // P stands for Pigeon and F for Foreign Host; this queue stores messages at P which are destined for F 
int size = 0;  /* number of full elements */
int front,rear=0;  /* queue end indices*/


int* Queue_PH; // P stands for Pigeon and F for Foreign Host; this queue stores messages at P which are destined for H 
int size1 = 0;  /* number of full elements */
int front1,rear1=0;  /* queue end indices*/



int location; // varies from 0 to "distance" for the time being: H is at 0 and F is at "distance"; this variable is updated only by mobility module
int Direction; // 0 means pigeon is not moving, 1 means moving towards H, -1 means moving towards F // this flag is updated only by Scheduler 


int pickedup; // number of messages currently picked up from H; updted by MPM module

int delivered; // number of messages currently delivered to F; updted by MDM module


int pickedup1; // number of messages currently picked up from F; updted by MPM module

int delivered1; // number of messages currently delivered to H; updted by MDM module


int ConnectableH; // 1 means H is connectable and 0 means not ; updated by Connectivity module C which is called only by MPM. 


int ConnectableF; // 1 means F is connectable and 0 means not ; updated by Connectivity module C which is called only by MDM. 

int velocity=20; // default pigeon speed is 20 units; we assume linear speed
int distance=100; // default distance between home and foreign is 100 units

time_t t;// for seed of the random number generator



main(int argc, char** argv)  
{
  pthread_t M_thread; // mobility module 
  pthread_t S_thread; //scheduler module

  void *M();
  void *S();
  int genDelayHome=1; // default generation delay in each host (in seconds)
  int genDelayForeign=1; // default generation delay in each host (in seconds)

  if ((argc !=7 && argc != 11))    /* Test for correct number of arguments */
    {
       fprintf(stderr, "Usage: %s <HomeIP> <HomeMT_Port> <HomeMR_Port> <ForeignIP> <ForeignMT_Port> <ForeignMR_Port> [<travelDistancc-in-integer> <velocity-in-integer> <messageGenHomeDelay-in-integer> <messageGenDelayForeign-in-integer>]\n",
               argv[0]);
       exit(1);
    }


    strcpy(HomeIP, argv[1]);             /* First arg: home IP address*/

    HomeMT_Port = atoi(argv[2]); /* home message port */

    HomeMR_Port = atoi(argv[3]); /* home message port */

    strcpy(foreignIP, argv[4]);             /* 4th arg: foreign IP address*/

    foreignMT_Port = atoi(argv[5]); /* foreign message  transfer port */
    foreignMR_Port = atoi(argv[6]); /* foreign message  receiving port */

    SIZE_PF=50; // default queue size

    SIZE_PH=50; // default queue size

    if(argc == 11) 
    {
      distance = atoi(argv[7]);             /* 5th arg: distance*/

      velocity = atoi(argv[8]); /* 6th arg: velocity */

      genDelayHome=atoi(argv[9]); /*7th arg: mean message generation delay */
      genDelayForeign=atoi(argv[10]); /*7th arg: mean message generation delay */

      SIZE_PF = (int)(((2*((float)distance / (float)velocity))/(float)genDelayHome)*5);
                                      // 5 is the safety factor in above exp
      SIZE_PH = (int)(((2*((float)distance / (float)velocity))/(float)genDelayForeign)*5);
				      // 5 is the safety factor in above exp
      
      if(SIZE_PF < 10)
        SIZE_PF=10;

      
      if(SIZE_PH < 10)
        SIZE_PH=10;
    }


    printf("queue_PF size in pigeon is %d \n", SIZE_PF);
 
    printf("queue_PH size in pigeon is %d \n", SIZE_PH);
 
    Queue_PF=(int*)malloc(SIZE_PF*sizeof(int)); // getting stoarge space of the queue

    Queue_PH=(int*)malloc(SIZE_PH*sizeof(int)); // getting stoarge space of the queue

    // initialize global variables which represent the state of the Pigeon
 
    location = 0; // P is at 0 at the starting point
    Direction = 0; // 0 means pigeon is not moving 


    ConnectableH = 0; // 0 means H is not connectable now
    ConnectableF=0;


    pthread_create(&M_thread,NULL,M,NULL);
    pthread_create(&S_thread,NULL,S,NULL);


    pthread_join(S_thread,NULL);

}



// following two functions take care of enqueue and dequeue of Queue_PF

void add_queue(int i){
  Queue_PF[rear] = i;
  rear = (rear+1) % SIZE_PF;
  size++;
}

int get_queue(){
  int v;
  v = Queue_PF[front];
  front= (front+1) % SIZE_PF;
  size--;
  return v ;
}

// following two functions take care of enqueue and dequeue of Queue_PH

void add_queue1(int i){
  Queue_PH[rear1] = i;
  rear1 = (rear1+1) % SIZE_PH;
  size1++;
}

int get_queue1(){
  int v;
  v = Queue_PH[front1];
  front1= (front1+1) % SIZE_PH;
  size1--;
  return v ;
}


// This is the body of the mobility module; 

void *M() 
{

    int numRcvd, temp; 

    int templocation=0;

    int upH=0, upF=0;

  while(1)
  {
    sleep(1);
    templocation=location;
    if(Direction==1 && location < distance)
    {
      if(upH==1)
        {
          system("iptables -A INPUT -s 10.0.1.2 -j DROP");
          upH=0;
        }
      templocation = location + velocity;
      
      if(templocation > distance) // handling the case where location is beyond home
        templocation=distance;

      location = templocation;

      if(location==distance && upF==0)
       {
          upF=1;
          
          system("iptables -D INPUT -s 10.0.2.2 -j DROP");
       }
    }
    else if (Direction==-1 && location > 0)
    {
      
      if(upF==1)
        {
          system("iptables -A INPUT -s 10.0.2.2 -j DROP");
          upF=0;
        }
       
      templocation = location - velocity;
      if(templocation < 0)  // handling the case when location is negative(beyond foreign)
        templocation = 0;
      location = templocation;

      
      if(location==0 && upH==0)
       {
          upH=1;
          
          system("iptables -D INPUT -s 10.0.1.2 -j DROP");
       }
    }

    printf("from M: location = %d \n",location);

    
    printf("from M: number of messages in pigeon PF queue = %d \n",size);

    printf("from M: number of messages in pigeon PH queue = %d \n",size1);
  }
  
  pthread_exit(NULL);

}

// Function readable_timeo() uses select() to see if file descriptor fd is readable. It keeps on checking if the file descriptor fd is readable. If fd is not readable for longer than a duration, a timeout occurs.

 int readable_timeo(int fd, int sec) 
 {
   fd_set rset;
   struct timeval tv;
   
   FD_ZERO(&rset);
   FD_SET(fd, &rset);
  
   tv.tv_sec = sec;
   tv.tv_usec = 0;

   return(select(fd + 1, &rset, NULL, NULL, &tv)); 
   // > 0 if fd is readable; = 0 if timeout occurs; > 0 if fd is redable   
 }



// This is the body of the connectivity module C; it includes a UDP ping client to check the connectivity with a host node
int C( char* echoServIP, unsigned short echoServPort) 
{

    int sock;                        /* Socket descriptor */
    struct sockaddr_in echoServAddr; /* ping (also called echo) server address */
    struct sockaddr_in fromAddr;     /* Source address of ping */
    unsigned int fromSize;           /* In-out of address size for recvfrom() */
    
    
    char *echoString = "ping";         /* String to send to ping server */
    char echoBuffer[ECHOMAX+1];      /* Buffer for echo string */
    int echoStringLen;               /* Length of string to echo */
    int respStringLen;               /* Size of received datagram */

    int tries=0;
    int ret;

    if ((echoStringLen = strlen(echoString)) > ECHOMAX)
        DieWithError("Echo word too long");


   /* Create a best-effort datagram socket using UDP */
    if ((sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
        DieWithError("socket() failed");

    /* Construct the server address structure */
    memset(&echoServAddr, 0, sizeof(echoServAddr));    /* Zero out structure */
    echoServAddr.sin_family = AF_INET;
    echoServAddr.sin_addr.s_addr = inet_addr(echoServIP);  /* Server IP address */
    echoServAddr.sin_port = htons(echoServPort);       /* Server port */

    /* Send the string to the server */
    if (sendto(sock, echoString, echoStringLen, 0, (struct sockaddr *)
               &echoServAddr, sizeof(echoServAddr)) != echoStringLen)
        DieWithError("from C(): sendto() sent a different number of bytes than expected");
    
    /* Get a response */
    fromSize = sizeof(fromAddr);


    while (tries < MAXTRIES)
     { 
       ret=readable_timeo(sock, TIMEOUT_SECS);
       if(ret  == 0) {
           printf("***UDP Socket timed out for %d time \n", (tries+1));

           if (sendto(sock, echoString, echoStringLen, 0, (struct sockaddr *)
                            &echoServAddr, sizeof(echoServAddr)) != echoStringLen)
                    DieWithError("from c(): sendto() failed");
           
           tries++;

       } 
       else if(ret > 0) {
           respStringLen = recvfrom(sock, echoBuffer, ECHOMAX, 0,
           (struct sockaddr *) &fromAddr, &fromSize);
           break;       

       } 
       else {
           printf("C(): error from select() call \n");
           close(sock);
           return 0;
       }
       

     } // while()
    
    if(tries==MAXTRIES)
      { 
         printf("from C(): MAXTRIES for UDP ping reached \n"); 
         close(sock); 
         return 0;
      }

       


    /* null-terminate the received data */
    echoBuffer[respStringLen] = '\0';
    printf("from C(): Received: %s\n", echoBuffer);    /* Print the received data */
    
    close(sock);


    return 1;

}




void *S()
{
  void MPM(char*, unsigned short);
  void MDM(char*, unsigned short);

 
  while(1)
  { 
    pickedup=0; // initializing current number of picked up messages

    delivered=0; // initializing current number of delivered messages

    pickedup1=0; // initializing current number of picked up messages

    delivered1=0; // initializing current number of delivered messages

    printf("from S: scheduler loop starts\n");
    while(location < distance)
    {
       if(Direction !=1)     
          Direction=1;
    }

    printf("from S: pigeon arrives at Foreign \n");
      Direction=0;   
     MDM(foreignIP, foreignMR_Port); // commented for testing
     MPM(foreignIP, foreignMT_Port); // commented for testing
     
      printf("Scheduler:after pickup and delivery at Foreign Host\n");

    while(location > 0)
    {
      if(Direction !=-1)
        Direction=-1;
    }

    
    printf("from S: pigeon arrives at Home Host\n");

      Direction=0;   
      MDM(HomeIP, HomeMR_Port);
      MPM(HomeIP, HomeMT_Port);
      
      printf("Scheduler:after pickup and delivery at Home Host\n");

    printf("Scheduler one loop ends \n");
  
  }//while(1)    
  pthread_exit(NULL);
 
}

 
void MPM(char* ServIP, unsigned short ServPort)
{

    
    int i = 0;

    int sock;                        /* Socket descriptor */
    struct sockaddr_in ServAddr; /* server address */
    int number, numRcvd, temp; 
    int Connectable=0; // flag to indicate connection status of ServIP
    number = 0; // number of messages to pick up in one visit


    /* Construct the server address structure */
    memset(&ServAddr, 0, sizeof(ServAddr));     /* Zero out structure */
    ServAddr.sin_family      = AF_INET;             /* Internet address family */
    ServAddr.sin_addr.s_addr = inet_addr(ServIP);   /* Server IP address */
    ServAddr.sin_port        = htons(ServPort); /* Server port */

    Connectable = C(ServIP, PING_PORT);//calling the connectivity module to update Connectable  
    
   printf("from MPM: for ServIP=%s Connectable=%d \n", ServIP, Connectable);
 
    if(Connectable==0)
      {

       printf("*********MPM: ServIP=%s (Port = PING_PORT) is not Connectable: will try later \n",ServIP);
       goto Last;
      }

     // we are here only when Connectable==1

    /* Create a reliable, stream socket using TCP */
    if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
        DieWithError("socket() failed");


    /* Establish the connection to the server */
    if (connect(sock, (struct sockaddr *) &ServAddr, sizeof(ServAddr)) < 0)
        { 
           printf("connect() failed; will try after some time \n");
           
           goto Last;
           
        }


 /*receive the number of numbers to be received from the server*/
    if (recv(sock, &number, sizeof(int), 0) != sizeof(int))
        DieWithError("recv() recvd a different number of bytes than expected");

 // The following if-else handles the condition when pigeon queue does (or not) have space to pick up all messages currently available in host; accordingly the pigeon lets the host know the final number by send().    

    if(!strcmp(ServIP, HomeIP)) // when ServIP is HomeIP
    {

       if(size+number <= SIZE_PF)  

         {
       
            if (send(sock, &number, sizeof(int), 0) != sizeof(int))
                DieWithError("send() recvd a different number of bytes than expected");
       

         } else {
       
               number = SIZE_PF - size; // CHECK 
               if (send(sock, &number, sizeof(int), 0) != sizeof(int))
                DieWithError("send() recvd a different number of bytes than expected");
        }

       printf("%d numbers to Receive from %s host:\n", number, ServIP);               
    
       while (pickedup < number)
        {
           if ((temp = recv(sock, &numRcvd, sizeof(int), 0)) <= 0)
             DieWithError("recv() failed or connection closed prematurely");
           pickedup+= 1;   /* Keep tally of total numbers */
           printf("number received from host (%s) is %d \n", ServIP, numRcvd);  /* Print the number */

           pthread_mutex_lock(&region_mutex);
           if (size == SIZE_PF) {

            printf("*******pigeon PF queue is full*****\n");
            pthread_cond_wait(&space_available,&region_mutex);

        }
        add_queue(numRcvd);
        pthread_cond_signal(&data_available);
        pthread_mutex_unlock(&region_mutex);

       }//while()

     } else if(!strcmp(ServIP, foreignIP)) // when ServIP is foreignIP

     {
   
       if(size1+number <= SIZE_PH)  

        {
       
          if (send(sock, &number, sizeof(int), 0) != sizeof(int))
            DieWithError("send() recvd a different number of bytes than expected");
       

        } else {
       
            number = SIZE_PH - size1; // CHECK 
            if (send(sock, &number, sizeof(int), 0) != sizeof(int))
             DieWithError("send() recvd a different number of bytes than expected");
        }

        printf("%d numbers to Receive from %s host:\n", number, ServIP);               
    
        while (pickedup1 < number)
         {
           if ((temp = recv(sock, &numRcvd, sizeof(int), 0)) <= 0)
            DieWithError("recv() failed or connection closed prematurely");
           pickedup1+= 1;   /* Keep tally of total numbers */
           printf("number received from host(%s) is %d \n", ServIP, numRcvd);  /* Print the number */

          pthread_mutex_lock(&region_mutex1);
          if (size1 == SIZE_PH) {

             printf("*******pigeon PH queue is full*****\n");
             pthread_cond_wait(&space_available1,&region_mutex1);

          }
          add_queue1(numRcvd);
          pthread_cond_signal(&data_available1);
          pthread_mutex_unlock(&region_mutex1);

        }// while
    } // else-if

    close(sock);

    Last:
    printf("end of MPM \n");  
  
return;
}


void MDM(char* ServIP, unsigned short ServPort)
{
    
    int v;
    int j;
    int sock;                        /* Socket descriptor */
    struct sockaddr_in ServAddr; /* server address */
    int number, numRcvd, temp; 
    int Connectable=0; // flag to indicate connection status of ServIP
    
    number = 0; // initialize the number of messages to deliver in one visit


    /* Construct the server address structure */
    memset(&ServAddr, 0, sizeof(ServAddr));     /* Zero out structure */
    ServAddr.sin_family      = AF_INET;             /* Internet address family */
    ServAddr.sin_addr.s_addr = inet_addr(ServIP);   /* Server IP address */
    ServAddr.sin_port        = htons(ServPort); /* Server port */



    
    Connectable = C(ServIP, PING_PORT); //calling the connectivity module to update Connectable 
    
    if(Connectable==0)
      {

       printf("*********MDM: %s is not Connectable: will try later \n", ServIP);
       goto Last;
      }

     // we are here only when Connectable==1

    /* Create a reliable, stream socket using TCP */
    if ((sock = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
        DieWithError("socket() failed");


    /* Establish the connection to the server */
    if (connect(sock, (struct sockaddr *) &ServAddr, sizeof(ServAddr)) < 0)
        { 
           printf("connect() failed; will try after some time \n");
           goto Last;
           
        }

    if(!strcmp(ServIP, foreignIP)) // if ServIP is foreign Host
       {
          number = size; // We need to check this for atomicity

         /* Send to F the number of numbers to be delivered to F : ROY */
         if (send(sock, &number, sizeof(int), 0) != sizeof(int))
           DieWithError("send() sent a different number of bytes than expected");

         printf("%d messages to deliver to Foreign Host\n", number);

    
         for(j=0;j<number;j++) 
           {


             pthread_mutex_lock(&region_mutex);
             if (size == 0) {
               printf("****MDM:the PF queue is empty****\n");
               pthread_cond_wait(&data_available,&region_mutex); 
             } 
             v = get_queue();
             pthread_cond_signal(&space_available);
             pthread_mutex_unlock(&region_mutex);
             printf("number to deliver to %s is %d  \n", ServIP, v);

   
            if ((temp = send(sock, &v, sizeof(int), 0)) <= 0)
               DieWithError("send() failed or connection closed prematurely");
            delivered+= 1;   /* Keep tally of total numbers */
            printf("number delivered  to foreign host is %d \n", v);  /* Print the number */

           } // for(j=0;...)

      } else if(!strcmp(ServIP, HomeIP)) // if ServIP is Home Host
    
      {
          number = size1; // We need to check this for atomicity

          /* Send to H the number of numbers to be delivered to H */
          if (send(sock, &number, sizeof(int), 0) != sizeof(int))
            DieWithError("send() sent a different number of bytes than expected");

          printf("%d messages to deliver to Home Host\n", number);

    
          for(j=0;j<number;j++) 
           {

             pthread_mutex_lock(&region_mutex1);
             if (size1 == 0) {
                printf("****MDM:the PH queue is empty****\n");
                pthread_cond_wait(&data_available1,&region_mutex1); 
             } 
             v = get_queue1();
             pthread_cond_signal(&space_available1);
             pthread_mutex_unlock(&region_mutex1);
             printf("number to deliver is %d  \n",v);

   
            if ((temp = send(sock, &v, sizeof(int), 0)) <= 0)
              DieWithError("send() failed or connection closed prematurely");
            delivered1+= 1;   /* Keep tally of total numbers */
            printf("number delivered  to host(%s) is %d \n", ServIP, v);  /* Print the number */

          } // for(j=0;...)

    } // else-if

    close(sock);

  Last:  

    printf("end of while loop in MDM \n");  


return;
}
 

